"""Count splits that occur in SortedList._expand with varying loads.

NOTE: Benchmarking performance at scale eventually used sampling rather than
what is described below. Do not use this method for benchmarking. This file is
kept for SortedListWithSplits which presents an example for recording the
number of sublist splits that occur.

Compare three load styles:

1. Low - Simply initializing a SortedList creates each sublist with length
equal to `load`.

2. Normal Random - See `init_sorted_list` below. Each sublist is given a length
chosen at random from a normal distribution with mean 1500 and standard
deviation 100. See `plot_lengths_histogram.py` for a visualization of the
sublist length and motivation for the normal distribution parameters.

3. Added Random - Each sublist is initialized by simply adding random values to
the list. This one is the most realistic but expensive to compute. It's also
biased by the number of values added. Because SortedList amortizes some costs,
this metric is prone to under or over-estimating impact.

Sample output: LIMIT = int(1e6), RATIO = 1

Low [504, 490, 487, 503, 488, 494, 496, 502, 493, 486]
Normal Random [666, 667, 662, 666, 670, 668, 668, 666, 668, 666]
Added Random [685, 675, 689, 687, 699, 706, 691, 679, 672, 702]

Conclusion:

Normal Random is about 1-2% less than Added Random on average but remains
practical for use as an efficient approximation.

The above values were calculated with RATIO = 1. Increasing the ratio to 10
creates a significant divergence between Normal Random and Added Random. Sample
output with RATIO = 10:

Low [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
Normal Random [1, 2, 0, 1, 0, 1, 0, 1, 1, 2]
Added Random [307, 286, 290, 253, 279, 319, 283, 264, 278, 280]

The difference is due to the load of the sublists. Randomly adding a million
values positions the sublist lengths in a bimodal distribution split at 1000
and 2000. This is nearly the worst-case scenario for measurements.

After trying a number of settings, RATIO = 3 appears reliable and useful. This
approximates the available capacity:

    >>> random.seed(0)
    >>> sl = init_sorted_list(sc.SortedList(), int(1e6))
    >>> len(sl._lists) * sl._load * 2 - len(sl)
    334000

So after constructing a SortedList with sublist length given by a normal
distribution with mean 1,500 and standard deviation 100, we fill the remaining
empty capacity.

"""

from __future__ import print_function

import sortedcontainers as sc
import random

REPEAT = 5
LIMIT = int(1e6)
RATIO = 3


class SortedListWithSplits(sc.SortedList):
    "SortedList that counts splits that occur in _expand."
    
    def __init__(self, *args, **kwargs):
        self.splits = 0
        super(SortedListWithSplits, self).__init__(*args, **kwargs)

    def _expand(self, pos):
        if len(self._lists[pos]) > self._twice:
            self.splits += 1
        super(SortedListWithSplits, self)._expand(pos)


def init_sorted_list(sl, size):
    "Initialize a SortedList with normally distributed sublist lengths."
    sl.clear()

    total = 0

    while total < size:
        # count = random.randrange(sl._load, sl._twice)
        count = int(random.normalvariate(int(sl._load * 1.5), 100))
        count = min(count, sl._load * 2)
        count = max(count, sl._load)
        sl._lists.append(list(range(total, total + count)))
        total += count

    sl._len = sum(len(sublist) for sublist in sl._lists)
    sl._maxes[:] = [sublist[-1] for sublist in sl._lists]

    sl._check()

    while len(sl) > size:
        del sl[random.randrange(len(sl))]

    return sl


def fill(obj, count, limit):
    "Repeatedly add random values to the SortedList."
    for each in range(count):
        obj.add(random.randrange(limit))


if __name__ == '__main__':
    low_splits = []

    for each in range(REPEAT):
        sl = SortedListWithSplits(range(LIMIT))
        fill(sl, LIMIT / RATIO, LIMIT)
        low_splits.append(sl.splits)

    uniform_splits = []

    for each in range(REPEAT):
        sl = init_sorted_list(SortedListWithSplits(), LIMIT)
        fill(sl, LIMIT / RATIO, LIMIT)
        uniform_splits.append(sl.splits)

    rand_splits = []

    for each in range(REPEAT):
        sl = SortedListWithSplits()

        values = list(range(LIMIT))
        random.shuffle(values)
        for value in values:
            sl.add(value)

        sl.splits = 0

        fill(sl, LIMIT / RATIO, LIMIT)
        rand_splits.append(sl.splits)

    print('Low', low_splits)
    print('Normal Random', uniform_splits)
    print('Added Random', rand_splits)



"""
# Plotting Benchmark Results

## Usage

    usage: benchmark_plot.py [-h] [--test TEST] [--kind KIND] [--show] [--save]
                             filename name

    Plotting

    positional arguments:
      filename     path to file with benchmark data
      name         type name

    optional arguments:
      -h, --help   show this help message and exit
      --test TEST
      --kind KIND
      --show
      --save

"""

from __future__ import print_function

import argparse
import matplotlib
import matplotlib.pyplot as plt
from collections import OrderedDict

class TreeDict(OrderedDict):
    def __missing__(self, key):
        self[key] = value = TreeDict()
        return value

def order_kinds(kinds):
    for idx, kind in enumerate(kinds):
        if kind.startswith('Sorted'):
            del kinds[idx]
            kinds.insert(0, kind)
            break
    return kinds

def test_plot(test):
    ax = plt.gca()
    ax.grid(linestyle='dashed', linewidth=0.5)
    cmap = matplotlib.cm.get_cmap('Set1')
    colors = list(cmap.colors)
    colors[0], colors[1] = colors[1], colors[0]
    del colors[5]
    ax.set_prop_cycle('color', colors)
    kinds = args.kind or list(data[test])
    for order, kind in enumerate(kinds):
        kind_plot(test, kind, len(kinds) - order)
    plt.ylim(ymin=9e-7)
    plt.loglog()
    plt.title(args.name + ' Performance: ' + test)
    plt.ylabel('Seconds')
    plt.xlabel('List Size')
    plt.legend(kinds, loc=2)

def kind_plot(test, kind, zorder):
    sizes = sorted(data[test][kind].keys())
    # Timer isn't any better than micro-second resolution.
    yvalues = [max(1e-6, data[test][kind][size][5]) for size in sizes]
    plt.plot(sizes, yvalues, marker='s', zorder=zorder)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Plotting')
    parser.add_argument('filename', help='path to file with benchmark data')
    parser.add_argument('name', help='type name')
    parser.add_argument('--test', action='append', help='filter tests by name')
    parser.add_argument('--kind', action='append', help='filter types by name')
    parser.add_argument('--suffix', default='', help='suffix for output')
    parser.add_argument('--show', action='store_true', help='show results')
    parser.add_argument('--save', action='store_true', help='save results')
    args = parser.parse_args()

    text = open(args.filename).read()

    lines = text.splitlines()
    lines = [line.split() for line in lines]
    for line in lines:
        line[2] = int(line[2])
        line[3:] = map(float, line[3:])
    data = TreeDict()
    for line in lines:
        data[line[0]][line[1]][line[2]] = line

    tests = args.test or list(data)

    for test in tests:
        test_plot(test)

        if args.show:
            plt.show()

        if args.save:
            plt.savefig('{0}{1}-{2}.png'.format(args.name, args.suffix, test))

        plt.close()



"""Plot histogram of SortedList sublist lengths while items are added.

Script output is a video named `sublist-lengths.mp4`. The video displays a line
graph where the x-axis is sublist length and the y-axis is the percentage of
sublists with that length. Sublist lengths are divided into buckets as with a
histogram. A line fitting the histogram as a normal distribution is plotted
alongside in green.

A line graph was used to represent the histogram because it was easier to
animate.

Observations:

* Average sublist length cycles between `load` and `load * 2` with an
  exponentially increasing period.

* As the number of elements increases, the sublist length distribution can be
  approximated by a normal distribution. The fit is valid within 30% of the
  mean.

* At the end of the video, the normal distribution fitting the histogram with
  mean=1500 has standard deviation=102.94.

* Observing the mean and standard deviation over time shows that the standard
  deviation tends to increase at a given mean over time. In this way, the curve
  flattens as more values are added to it.

"""

from __future__ import division, print_function

import argparse
import itertools as it
import numpy as np
import scipy
import scipy.stats
import sortedcontainers as sc
from matplotlib import pyplot as plt
from matplotlib import animation
import random

DISPLAY_FIT = False

def background():
    "Plot the background of each animated frame."
    hist_line.set_data([], [])
    norm_line.set_data([], [])
    return hist_line, norm_line


def frame(num):
    "Draw frame."
    values.update(func() for func in it.repeat(random.random, LOAD))
    data = np.array([len(sublist) for sublist in values._lists])
    hist, bins = np.histogram(data, bins=BINS, normed=True)
    hist_line.set_data(bins[:-1], hist)
    norm = scipy.stats.norm
    params = norm.fit(data)
    norm_line.set_data(LENGTHS, norm.pdf(LENGTHS, *params))

    if DISPLAY_FIT:
        print('Normal', '%10.3f %10.3f' % params)

    return hist_line, norm_line


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--seed', type=int, default=0)
    parser.add_argument('--load', type=float, default=1e3)
    parser.add_argument('--show', action='store_true')
    parser.add_argument('--filename', default='sublist-lengths-add.mp4')
    parser.add_argument('--display-fit', action='store_true')

    args = parser.parse_args()

    random.seed(args.seed)

    LOAD = int(args.load)
    BINS = np.array(range(LOAD, LOAD * 2 + 1, LOAD // 20))
    LENGTHS = np.array(range(LOAD, LOAD * 2))
    DISPLAY_FIT = args.display_fit

    values = sc.SortedList(load=LOAD)

    # Initialize with random values.

    for func in it.repeat(random.random, LOAD):
        values.add(func())

    fig = plt.figure()
    bounds = (LOAD - 5 * LOAD // 100, LOAD * 2 - 5 * LOAD // 100)
    ax = plt.axes(xlim=bounds, ylim=(0, 0.01))
    hist_line, = ax.plot([], [])
    norm_line, = ax.plot([], [])

    animator = animation.FuncAnimation(
        fig, frame, init_func=background, frames=1000, interval=10, blit=True,
    )

    if args.show:
        plt.show()
    else:
        animator.save(
            args.filename, fps=20, extra_args=['-vcodec', 'libx264']
        )



"""Benchmark SortedContainers at scale.

## Local Results

$ pypy benchmark_scale.py --limit 1e9
@   Method    Size   Operations           Time       Ops/Sec     Ratio
@      add   1e+06        1e+04        0.01501    666045.025       nan
@      add   1e+07        1e+05        0.26612    375764.681     1.773
@      add   1e+08        1e+06        4.69080    213183.298     1.763
@      add   1e+09        1e+07       83.01831    120455.358     1.770
@      del   1e+06        1e+04        0.00827   1208897.485       nan
@      del   1e+07        1e+05        0.13309    751393.836     1.609
@      del   1e+08        1e+06        3.79143    263752.866     2.849
@      del   1e+09        1e+07      124.59184     80262.081     3.286

## High Memory Instance Results

Note: Requires ~128 GB of memory.
Note: Requires ~11 hrs to run.

$ nohup /home/grantj/PyPy27/bin/pypy benchmark_scale.py &
$ cat nohup.out
@   Method    Size     Ops           Time       Ops/Sec     Ratio
@      add   1e+06   1e+04        0.02133    468884.826       nan
@      add   1e+07   1e+05        0.38629    258872.924     1.811
@      add   1e+08   1e+06        6.20695    161109.825     1.607
@      add   1e+09   1e+07      120.24735     83161.919     1.937
@      add   1e+10   1e+08     2416.60713     41380.330     2.010
@      del   1e+06   1e+04        0.01791    558289.343       nan
@      del   1e+07   1e+05        0.26171    382097.449     1.461
@      del   1e+08   1e+06        6.11150    163626.036     2.335
@      del   1e+09   1e+07      171.58899     58278.798     2.808
@      del   1e+10   1e+08     5493.95076     18201.838     3.202

### CPU Info

Architecture:          x86_64
CPU op-mode(s):        32-bit, 64-bit
Byte Order:            Little Endian
CPU(s):                32
On-line CPU(s) list:   0-31
Thread(s) per core:    2
Core(s) per socket:    16
Socket(s):             1
NUMA node(s):          1
Vendor ID:             GenuineIntel
CPU family:            6
Model:                 63
Model name:            Intel(R) Xeon(R) CPU @ 2.30GHz
Stepping:              0
CPU MHz:               2299.998
BogoMIPS:              4599.99
Hypervisor vendor:     KVM
Virtualization type:   full
L1d cache:             32K
L1i cache:             32K
L2 cache:              256K
L3 cache:              46080K
NUMA node0 CPU(s):     0-31

"""
from __future__ import print_function

import argparse
import collections as co
import functools as ft
import gc
import itertools as it
import math
import random
import sortedcontainers as sc
import sys
import time


CHECK = False


def iter_with_description(iterable, description=''):
    "Placeholder iterator function with ignored description parameter."
    return iter(iterable)


PROGRESS = iter_with_description


def init_sorted_list(sl, size, moment=5, fraction=0.1):
    """Initialize SortedList with normally distributed sublist lengths.

    The mean of the normal distribution is given by:

        mu = load * (1.0 + moment / 10.0)

    And the standard deviation of the normal distribution is given by:

        sigma = load * fraction

    For a visualization of positive and negative moments see:

        * plot_lengths_histogram.py
        * plot_lengths_histogram_delitem.py

    :param SortedList sl: SortedList to initialize
    :param int size: size of the resulting SortedList
    :param int moment: number between -5 and 9 inclusive
    :param float fraction: fraction of load to be used as standard deviation
    :return: initialized sorted list

    """
    assert -5 <= moment <= 9
    assert 0 < fraction

    sl.clear()

    load = sl._load
    half = sl._load >> 1
    twice = sl._load << 1
    mu = load * (1.0 + moment / 10.0)
    sigma = load * fraction
    total = 0

    class WhileIterator(object):
        "Convert for-loop to while-loop with length estimate."
        def __iter__(self):
            while total < size:
                yield True
        def __len__(self):
            if moment < 0:
                return size / half
            else:
                return size / load

    for each in PROGRESS(WhileIterator(), 'init-sub'):
        count = int(random.normalvariate(mu, sigma))

        if moment >= 0:
            if count < load:
                count += load
            elif count > twice:
                count -= load

            count = min(count, twice)
            count = max(count, load)
        else:
            if count < half:
                count += half
            elif count > load:
                count -= half

            count = min(count, load)
            count = max(count, half)

        sl._lists.append(list(xrange(total, total + count)))
        total += count

    sl._len = sum(len(sublist) for sublist in sl._lists)
    sl._maxes[:] = [sublist[-1] for sublist in sl._lists]

    for each in PROGRESS(xrange(len(sl) - size), 'init-del'):
        del sl[random.randrange(len(sl))]

    del sl._index[:]

    if CHECK: sl._check()

    assert len(sl) == size

    return sl


def timeit(func):
    "Decorator to time function calls."
    @ft.wraps(func)
    def wrapper(*args, **kwargs):
        "Return timed duration of function call. Ignores function result."
        start = time.clock()
        result = func(*args, **kwargs)
        end = time.clock()
        return (end - start)
    return wrapper


@timeit
def add(obj, numbers):
    "Repeatedly add number from numbers to sorted list."
    for number in PROGRESS(numbers, 'add'):
        obj.add(number)


@timeit
def delitem(obj, indices):
    "Repeatedly delete values from sorted list by index in indices."
    for index in PROGRESS(indices, 'del'):
        del obj[index]


def randvalues(limit, fraction=0.001):
    "Return fraction of limit random values between 0 and limit."
    iterable = PROGRESS(xrange(int(limit * fraction)), 'randvalues')
    return [random.randrange(limit) for each in iterable]


def randindices(limit, fraction=0.002):
    "Return fraction of limit random indices counting down from limit."
    stop = limit - int(limit * fraction)
    iterable = PROGRESS(xrange(limit, stop, -1), 'randindices')
    return [random.randrange(length) for length in iterable]


def benchmark_add(start, limit, times):
    """Benchmark sorted list add method.

    Start and limit are an inclusive range of magnitudes.

    The load of the sorted list is the cube root of the size.

    Measurements are made by sampling performance at each "moment" of a sorted
    list while items are added to it. See `init_sorted_list` for how "moment"
    is used.

    """
    for exponent in xrange(start, limit + 1):
        timings = []
        count = 10 ** exponent
        sl = sc.SortedList(load=int(count ** (1.0 / 3)))

        for attempt in xrange(times):
            subtimings = []

            for moment in xrange(10):
                values = randvalues(count)
                init_sorted_list(sl, count, moment)
                gc.collect()
                subtiming = add(sl, values)
                subtimings.append(subtiming)

            timing = sum(subtimings)
            timings.append(timing)

        display('add', timings, count)


def benchmark_del(start, limit, times):
    """Benchmark sorted list delitem method.

    Start and limit are an inclusive range of magnitudes.

    The load of the sorted list is the square root of the size.

    Measurements are made by sampling performance at each "moment" of a sorted
    list while items are deleted from it. See `init_sorted_list` for how
    "moment" is used.

    """
    for exponent in xrange(start, limit + 1):
        timings = []
        count = 10 ** exponent
        sl = sc.SortedList(load=int(count ** (1.0 / 3))) # 2)))

        for attempt in xrange(times):
            subtimings = []

            for moment in xrange(-5, 0):
                indices = randindices(count)
                init_sorted_list(sl, count, moment)
                gc.collect()
                subtiming = delitem(sl, indices)
                subtimings.append(subtiming)

            timing = sum(subtimings)
            timings.append(timing)

        display('del', timings, count)


def display(name, times, size, last=['', 0]):
    "Display performance summary with ratio of ops/sec."

    times.sort()
    median_time = times[len(times) / 2]
    operations = size / 100
    ops_sec = operations / median_time

    last_name, last_ops_sec = last
    ratio = last_ops_sec / ops_sec if name == last_name else float('nan')
    last[0], last[1] = name, ops_sec

    template = '@%9s   %.0e   %.0e %14.5f  %12.3f    %6.3f'
    print(template % (name, size, operations, median_time, ops_sec, ratio))
    sys.stdout.flush()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--start', default=1e6, type=float)
    parser.add_argument('--limit', default=1e10, type=float)
    parser.add_argument('--seed', default=0, type=int)
    parser.add_argument('--times', default=5, type=int)
    parser.add_argument('--funcs', default='all', choices=['all', 'add', 'del'])
    parser.add_argument('--progress', action='store_true')
    parser.add_argument('--check', action='store_true')

    args = parser.parse_args()

    random.seed(args.seed)

    if args.progress:
        import tqdm
        PROGRESS = tqdm.tqdm

    CHECK = args.check

    start = int(math.log10(args.start))
    limit = int(math.log10(args.limit))

    template = '@%9s %7s %7s %14s  %12s %9s'
    header = 'Method', 'Size', 'Ops', 'Time', 'Ops/Sec', 'Ratio'
    print(template % header)
    sys.stdout.flush()

    if args.funcs == 'all':
        benchmarks = [benchmark_add, benchmark_del]
    elif args.funcs == 'add':
        benchmarks = [benchmark_add]
    elif args.funcs == 'del':
        benchmarks = [benchmark_del]
    else:
        raise ValueError(args.funcs)

    for benchmark in benchmarks:
        benchmark(start, limit, args.times)



# -*- coding: utf-8 -*-

from sys import hexversion

import random
from .context import sortedcontainers
from sortedcontainers import SortedList
from itertools import chain
import pytest

if hexversion < 0x03000000:
    from itertools import izip as zip
    range = xrange

def test_init():
    slt = SortedList()
    assert slt.key is None
    slt._check()

    slt = SortedList()
    slt._reset(10000)
    assert slt._load == 10000
    slt._check()

    slt = SortedList(range(10000))
    assert all(tup[0] == tup[1] for tup in zip(slt, range(10000)))

    slt.clear()
    assert slt._len == 0
    assert slt._maxes == []
    assert slt._lists == []
    slt._check()

def test_add():
    random.seed(0)
    slt = SortedList()
    for val in range(1000):
        slt.add(val)
        slt._check()

    slt = SortedList()
    for val in range(1000, 0, -1):
        slt.add(val)
        slt._check()

    slt = SortedList()
    for val in range(1000):
        slt.add(random.random())
        slt._check()

def test_update():
    slt = SortedList()

    slt.update(range(1000))
    assert len(slt) == 1000
    slt._check()

    slt.update(range(100))
    assert len(slt) == 1100
    slt._check()

    slt.update(range(10000))
    assert len(slt) == 11100
    slt._check()

    values = sorted(chain(range(1000), range(100), range(10000)))
    assert all(tup[0] == tup[1] for tup in zip(slt, values))

def test_contains():
    slt = SortedList()
    assert 0 not in slt

    slt.update(range(10000))

    for val in range(10000):
        assert val in slt

    assert 10000 not in slt

    slt._check()

def test_discard():
    slt = SortedList()

    assert slt.discard(0) == None
    assert len(slt) == 0
    slt._check()

    slt = SortedList([1, 2, 2, 2, 3, 3, 5])
    slt._reset(4)

    slt.discard(6)
    slt._check()
    slt.discard(4)
    slt._check()
    slt.discard(2)
    slt._check()

    assert all(tup[0] == tup[1] for tup in zip(slt, [1, 2, 2, 3, 3, 5]))

def test_remove():
    slt = SortedList()

    assert slt.discard(0) == None
    assert len(slt) == 0
    slt._check()

    slt = SortedList([1, 2, 2, 2, 3, 3, 5])
    slt._reset(4)

    slt.remove(2)
    slt._check()

    assert all(tup[0] == tup[1] for tup in zip(slt, [1, 2, 2, 3, 3, 5]))

def test_remove_valueerror1():
    slt = SortedList()
    with pytest.raises(ValueError):
        slt.remove(0)

def test_remove_valueerror2():
    slt = SortedList(range(100))
    slt._reset(10)
    with pytest.raises(ValueError):
        slt.remove(100)

def test_remove_valueerror3():
    slt = SortedList([1, 2, 2, 2, 3, 3, 5])
    with pytest.raises(ValueError):
        slt.remove(4)

def test_delete():
    slt = SortedList(range(20))
    slt._reset(4)
    slt._check()
    for val in range(20):
        slt.remove(val)
        slt._check()
    assert len(slt) == 0
    assert slt._maxes == []
    assert slt._lists == []

def test_getitem():
    random.seed(0)
    slt = SortedList()
    slt._reset(17)

    lst = list()

    for rpt in range(100):
        val = random.random()
        slt.add(val)
        lst.append(val)

    lst.sort()

    assert all(slt[idx] == lst[idx] for idx in range(100))
    assert all(slt[idx - 99] == lst[idx - 99] for idx in range(100))

def test_getitem_slice():
    random.seed(0)
    slt = SortedList()
    slt._reset(17)

    lst = list()

    for rpt in range(100):
        val = random.random()
        slt.add(val)
        lst.append(val)

    lst.sort()

    assert all(slt[start:] == lst[start:]
               for start in [-75, -25, 0, 25, 75])

    assert all(slt[:stop] == lst[:stop]
               for stop in [-75, -25, 0, 25, 75])

    assert all(slt[::step] == lst[::step]
               for step in [-5, -1, 1, 5])

    assert all(slt[start:stop] == lst[start:stop]
               for start in [-75, -25, 0, 25, 75]
               for stop in [-75, -25, 0, 25, 75])

    assert all(slt[:stop:step] == lst[:stop:step]
               for stop in [-75, -25, 0, 25, 75]
               for step in [-5, -1, 1, 5])

    assert all(slt[start::step] == lst[start::step]
               for start in [-75, -25, 0, 25, 75]
               for step in [-5, -1, 1, 5])

    assert all(slt[start:stop:step] == lst[start:stop:step]
               for start in [-75, -25, 0, 25, 75]
               for stop in [-75, -25, 0, 25, 75]
               for step in [-5, -1, 1, 5])

def test_getitem_slice_big():
    slt = SortedList(range(4))
    lst = list(range(4))

    itr = ((start, stop, step)
           for start in [-6, -4, -2, 0, 2, 4, 6]
           for stop in [-6, -4, -2, 0, 2, 4, 6]
           for step in [-3, -2, -1, 1, 2, 3])

    for start, stop, step in itr:
        assert slt[start:stop:step] == lst[start:stop:step]

def test_getitem_slicezero():
    slt = SortedList(range(100))
    slt._reset(17)
    with pytest.raises(ValueError):
        slt[::0]

def test_getitem_indexerror1():
    slt = SortedList()
    with pytest.raises(IndexError):
        slt[5]

def test_getitem_indexerror2():
    slt = SortedList(range(100))
    with pytest.raises(IndexError):
        slt[200]

def test_getitem_indexerror3():
    slt = SortedList(range(100))
    with pytest.raises(IndexError):
        slt[-101]

def test_delitem():
    random.seed(0)

    slt = SortedList(range(100))
    slt._reset(17)
    while len(slt) > 0:
        pos = random.randrange(len(slt))
        del slt[pos]
        slt._check()

    slt = SortedList(range(100))
    slt._reset(17)
    del slt[:]
    assert len(slt) == 0
    slt._check()

def test_delitem_slice():
    slt = SortedList(range(100))
    slt._reset(17)
    del slt[10:40:1]
    del slt[10:40:-1]
    del slt[10:40:2]
    del slt[10:40:-2]

def test_iter():
    slt = SortedList(range(10000))
    itr = iter(slt)
    assert all(tup[0] == tup[1] for tup in zip(range(10000), itr))

def test_reversed():
    slt = SortedList(range(10000))
    rev = reversed(slt)
    assert all(tup[0] == tup[1] for tup in zip(range(9999, -1, -1), rev))

def test_reverse():
    slt = SortedList(range(10000))
    with pytest.raises(NotImplementedError):
        slt.reverse()

def test_islice():
    sl = SortedList()
    sl._reset(7)

    assert [] == list(sl.islice())

    values = list(range(53))
    sl.update(values)

    for start in range(53):
        for stop in range(53):
            assert list(sl.islice(start, stop)) == values[start:stop]

    for start in range(53):
        for stop in range(53):
            assert list(sl.islice(start, stop, reverse=True)) == values[start:stop][::-1]

    for start in range(53):
        assert list(sl.islice(start=start)) == values[start:]
        assert list(sl.islice(start=start, reverse=True)) == values[start:][::-1]

    for stop in range(53):
        assert list(sl.islice(stop=stop)) == values[:stop]
        assert list(sl.islice(stop=stop, reverse=True)) == values[:stop][::-1]

def test_irange():
    sl = SortedList()
    sl._reset(7)

    assert [] == list(sl.irange())

    values = list(range(53))
    sl.update(values)

    for start in range(53):
        for end in range(start, 53):
            assert list(sl.irange(start, end)) == values[start:(end + 1)]
            assert list(sl.irange(start, end, reverse=True)) == values[start:(end + 1)][::-1]

    for start in range(53):
        for end in range(start, 53):
            assert list(range(start, end)) == list(sl.irange(start, end, (True, False)))

    for start in range(53):
        for end in range(start, 53):
            assert list(range(start + 1, end + 1)) == list(sl.irange(start, end, (False, True)))

    for start in range(53):
        for end in range(start, 53):
            assert list(range(start + 1, end)) == list(sl.irange(start, end, (False, False)))

    for start in range(53):
        assert list(range(start, 53)) == list(sl.irange(start))

    for end in range(53):
        assert list(range(0, end)) == list(sl.irange(None, end, (True, False)))

    assert values == list(sl.irange(inclusive=(False, False)))

    assert [] == list(sl.irange(53))
    assert values == list(sl.irange(None, 53, (True, False)))

def test_len():
    slt = SortedList()

    for val in range(10000):
        slt.add(val)
        assert len(slt) == (val + 1)

def test_bisect_left():
    slt = SortedList()
    assert slt.bisect_left(0) == 0
    slt = SortedList(range(100))
    slt._reset(17)
    slt.update(range(100))
    slt._check()
    assert slt.bisect_left(50) == 100
    assert slt.bisect_left(200) == 200

def test_bisect():
    slt = SortedList()
    assert slt.bisect(10) == 0
    slt = SortedList(range(100))
    slt._reset(17)
    slt.update(range(100))
    slt._check()
    assert slt.bisect(10) == 22
    assert slt.bisect(200) == 200

def test_bisect_right():
    slt = SortedList()
    assert slt.bisect_right(10) == 0
    slt = SortedList(range(100))
    slt._reset(17)
    slt.update(range(100))
    slt._check()
    assert slt.bisect_right(10) == 22
    assert slt.bisect_right(200) == 200

def test_copy():
    alpha = SortedList(range(100))
    alpha._reset(7)
    beta = alpha.copy()
    alpha.add(100)
    assert len(alpha) == 101
    assert len(beta) == 100

def test_copy_copy():
    import copy
    alpha = SortedList(range(100))
    alpha._reset(7)
    beta = copy.copy(alpha)
    alpha.add(100)
    assert len(alpha) == 101
    assert len(beta) == 100

def test_count():
    slt = SortedList()
    slt._reset(7)

    assert slt.count(0) == 0

    for iii in range(100):
        for jjj in range(iii):
            slt.add(iii)
        slt._check()

    for iii in range(100):
        assert slt.count(iii) == iii

    assert slt.count(100) == 0

def test_pop():
    slt = SortedList(range(10))
    slt._reset(4)
    slt._check()
    assert slt.pop() == 9
    slt._check()
    assert slt.pop(0) == 0
    slt._check()
    assert slt.pop(-2) == 7
    slt._check()
    assert slt.pop(4) == 5
    slt._check()

def test_pop_indexerror1():
    slt = SortedList(range(10))
    slt._reset(4)
    with pytest.raises(IndexError):
        slt.pop(-11)

def test_pop_indexerror2():
    slt = SortedList(range(10))
    slt._reset(4)
    with pytest.raises(IndexError):
        slt.pop(10)

def test_pop_indexerror3():
    slt = SortedList()
    with pytest.raises(IndexError):
        slt.pop()

def test_index():
    slt = SortedList(range(100))
    slt._reset(17)

    for val in range(100):
        assert val == slt.index(val)

    assert slt.index(99, 0, 1000) == 99

    slt = SortedList((0 for rpt in range(100)))
    slt._reset(17)

    for start in range(100):
        for stop in range(start, 100):
            assert slt.index(0, start, stop + 1) == start

    for start in range(100):
        assert slt.index(0, -(100 - start)) == start

    assert slt.index(0, -1000) == 0

def test_index_valueerror1():
    slt = SortedList([0] * 10)
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(0, 10)

def test_index_valueerror2():
    slt = SortedList([0] * 10)
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(0, 0, -10)

def test_index_valueerror3():
    slt = SortedList([0] * 10)
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(0, 7, 3)

def test_index_valueerror4():
    slt = SortedList([0] * 10)
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(1)

def test_index_valueerror5():
    slt = SortedList()
    with pytest.raises(ValueError):
        slt.index(1)

def test_index_valueerror6():
    slt = SortedList(range(10))
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(3, 5)

def test_index_valueerror7():
    slt = SortedList([0] * 10 + [2] * 10)
    slt._reset(4)
    with pytest.raises(ValueError):
        slt.index(1, 0, 10)

def test_mul():
    this = SortedList(range(10))
    this._reset(4)
    that = this * 5
    this._check()
    that._check()
    assert this == list(range(10))
    assert that == sorted(list(range(10)) * 5)
    assert this != that

def test_imul():
    this = SortedList(range(10))
    this._reset(4)
    this *= 5
    this._check()
    assert this == sorted(list(range(10)) * 5)

def test_op_add():
    this = SortedList(range(10))
    this._reset(4)
    assert (this + this + this) == (this * 3)

    that = SortedList(range(10))
    that._reset(4)
    that += that
    that += that
    assert that == (this * 4)

def test_eq():
    this = SortedList(range(10))
    this._reset(4)
    assert this == list(range(10))
    assert this == tuple(range(10))
    assert not (this == list(range(9)))

def test_ne():
    this = SortedList(range(10))
    this._reset(4)
    assert this != list(range(9))
    assert this != tuple(range(11))
    assert this != [0, 1, 2, 3, 3, 5, 6, 7, 8, 9]
    assert this != (val for val in range(10))
    assert this != set()

def test_lt():
    this = SortedList(range(10, 15))
    this._reset(4)
    assert this < [10, 11, 13, 13, 14]
    assert this < [10, 11, 12, 13, 14, 15]
    assert this < [11]

def test_le():
    this = SortedList(range(10, 15))
    this._reset(4)
    assert this <= [10, 11, 12, 13, 14]
    assert this <= [10, 11, 12, 13, 14, 15]
    assert this <= [10, 11, 13, 13, 14]
    assert this <= [11]

def test_gt():
    this = SortedList(range(10, 15))
    this._reset(4)
    assert this > [10, 11, 11, 13, 14]
    assert this > [10, 11, 12, 13]
    assert this > [9]

def test_ge():
    this = SortedList(range(10, 15))
    this._reset(4)
    assert this >= [10, 11, 12, 13, 14]
    assert this >= [10, 11, 12, 13]
    assert this >= [10, 11, 11, 13, 14]
    assert this >= [9]

def test_repr():
    this = SortedList(range(10))
    this._reset(4)
    assert repr(this) == 'SortedList([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])'

def test_repr_recursion():
    this = SortedList([[1], [2], [3], [4]])
    this._lists[-1].append(this)
    assert repr(this) == 'SortedList([[1], [2], [3], [4], ...])'

def test_repr_subclass():
    class CustomSortedList(SortedList):
        pass
    this = CustomSortedList([1, 2, 3, 4])
    assert repr(this) == 'CustomSortedList([1, 2, 3, 4])'

def test_pickle():
    import pickle
    alpha = SortedList(range(10000))
    alpha._reset(500)
    beta = pickle.loads(pickle.dumps(alpha))
    assert alpha == beta
    assert alpha._load == beta._load

def test_build_index():
    slt = SortedList([0])
    slt._reset(4)
    slt._build_index()
    slt._check()

def test_check():
    slt = SortedList(range(10))
    slt._reset(4)
    slt._len = 5
    with pytest.raises(AssertionError):
        slt._check()
